/* ----------------------------------------------------------
		
		Psychedelic led lights with dsPic 30F2020 
		Marco Signorini - marco.signorini@libero.it
		
		File: main.c
	-----------------------------------------------------------
*/

#include "hardware.h"
#include "adc.h"
#include "filterbank.h"
#include "inbuffer.h"
#include "pwmeter.h"
#include "pwmdriver.h"
#include "mcp6s26.h"

/* Here are defined the configuration fuses */
_FOSCSEL(FRC_PLL);																				/* Set up for Internal Fast RC with PLL */
_FOSC(CSW_FSCM_OFF & FRC_HI_RANGE & OSC2_IO & HS_EC_DIS)
_FWDT(FWDTEN_OFF);																				/* Turn off the Watch-Dog Timer.  */
_FGS(CODE_PROT_OFF);																			/* Disable Code Protection */
_FBS(BWRP_WRPROTECT_OFF);

#ifndef _ICD_ENABLED_
	_FPOR(PWRT_128)																					/* Enable Power-up timer */
#else
	_FPOR(PWRT_OFF)																					/* Disable Power-up timer */
#endif

/* Global variables used by the main loop */
/* (defined here to be not included in the stack) */
fractional fInPower;
fractional fLowPower;
fractional fMidPower;
fractional fHiPower;
fractional *pOutBuffer;
unsigned char ucPowerScore;
unsigned char ucPowerScoreTmr;

/* ----------------------------------------------------------- */

#define LOWSTEP_INPOWER_VAL Q15(0.20)
#define HISTEP_INPOWER_VAL	Q15(0.40)

/* AGC Main routine */
void AGC_MainTask(unsigned char ucClippedSamples)
{
	fractional fValue = LOWSTEP_INPOWER_VAL;
	if (fInPower < fValue)
	{
		/* A too low buffer again... */
		ucPowerScore--;
		
		if (ucPowerScore == 0xFF)
		{
			/* Too much low buffers found... */
			ucPowerScoreTmr = 0;
			ucPowerScore = 0x80;
			
			/* ... it's time to increase external gain */
			PGA_IncGain();
		}
		
		return;
	}
	
	fValue = HISTEP_INPOWER_VAL;
	if ((fInPower > fValue) || (ucClippedSamples > SMPBUFFER/8))
	{
		/* A too high buffer again... */
		ucPowerScore++;
		
		if (ucPowerScore == 0)
		{
			/* Too much high buffers found... */
			ucPowerScoreTmr = 0;
			ucPowerScore = 0x80;
			
			/*... it's time to decrease external gain */
			PGA_DecGain();
			
			return;
		}
	}
	
	/* This is a good signal... */
	ucPowerScoreTmr++;

	/* .. and it's good for a long time */
	if (ucPowerScoreTmr == 0xFF)
	{
		ucPowerScoreTmr = 0x00;
		ucPowerScore = 0x80;
	}
}

/* ----------------------------------------------------------- */

/* This is the main task */
int main (void)
{
	/* Initialize the RS232 (debug) */
	DBG_INIT;
	
	/* Initialize global variables */
	fInPower = Q15(0.0);
	fLowPower = Q15(0.0);
	fMidPower = Q15(0.0);
	fHiPower = Q15(0.0);
	ucPowerScore = 0x80;
	ucPowerScoreTmr = 0;
	
	/* Initialize the external PGA */
	PGA_Init();

	/* Initialize the PWM driver */
	PWMD_Init();
	
	/* Initialize the input buffer */
	INBUF_Init();
	
	/* Initialize the filter bank */
	FLTBNK_Init();
	
	/* Initialize the AD converter */
	ADC_Init();
	
	/* Start the main loop */
	while (1)
	{
		/* There are enought samples to be processed? */
		if (INBUF_GetFilledSize() >= SMPBUFFER)
		{
			unsigned char ucPwmVal;

			/* Count clipped samples and evaluate input power */
 			ucPwmVal = PWMT_CountClipped(INBUF_GetBuffer());
			PWMT_EvalBuffer(INBUF_GetBuffer(), &fInPower);
			
			/* High pass to filter out the DC */
			pOutBuffer = FLTBNK_FilterDC();
			DBG_START;

			/* Dump AGC parameters */
			DBG_SENDVAL((void*)&fInPower, sizeof(fInPower));
			DBG_SENDCHAR(ucPwmVal);
			
			/* AGC updating routine */
			AGC_MainTask(ucPwmVal);
			
			/* Generate and evaluate low band */
			pOutBuffer = FLTBNK_FilterLow();
			ucPwmVal = CLAMP_UINT_TO_UCHAR(PWMT_EvalBuffer(pOutBuffer, &fLowPower)>>10);
			ucPwmVal = PWMD_SetLowValue(ucPwmVal);
			DBG_SENDVAL((void*)&fLowPower, sizeof(fLowPower));
			DBG_SENDCHAR(ucPwmVal);
			
			/* Generate and evaluate mid band */
			pOutBuffer = FLTBNK_FilterMid();
			ucPwmVal = CLAMP_UINT_TO_UCHAR(PWMT_EvalBuffer(pOutBuffer, &fMidPower)>>3);
			ucPwmVal = PWMD_SetMidValue(ucPwmVal);
			DBG_SENDVAL((void*)&fMidPower, sizeof(fMidPower));
			DBG_SENDCHAR(ucPwmVal);
			
			/* Generate and evaluate high band */
			pOutBuffer = FLTBNK_FilterHi();
			ucPwmVal = CLAMP_UINT_TO_UCHAR(PWMT_EvalBuffer(pOutBuffer, &fHiPower));
			ucPwmVal = PWMD_SetHiValue(ucPwmVal);
			DBG_SENDVAL((void*)&fHiPower, sizeof(fHiPower));
			DBG_SENDCHAR(ucPwmVal);

			/* Monitor the power score too */
			DBG_SENDCHAR(ucPowerScore);

			/* Monitor current gain factor */
			DBG_SENDCHAR(PGA_GetCurGain());

			/* Terminate the packet */
			DBG_END;

			/* Release this buffer */
			INBUF_Release();
		}
		
		/* Call the PWMDriver main loop routine */
		PWMD_MainLoop();
	}
}
