/* ----------------------------------------------------------
		
		Psychedelic led lights with dsPic 30F2020 
		Marco Signorini - marco.signorini@libero.it
		
		File: pwmdriver.c
	-----------------------------------------------------------
*/

#include "hardware.h"
#include "pwmdriver.h"

/* Definition of structure containing variables used by this module */
typedef struct _pwmdriverdata
{
	unsigned char ucLowPwmValue;		/* Required Pwm value for Low channel */
	unsigned char ucMidPwmValue;		/* Required Pwm value for Mid channel */
	unsigned char ucHiPwmValue;			/* Required Pwm value for Hi channel */
	
	unsigned char ucBrightTimer;		/* Main timer to be used to generate PWMs */
	unsigned char ucProcess;				/* Flag signalling that the timer value was changed */
} pwmdriverdata;

/* Allocation of local variables */
pwmdriverdata gtPWMDData;

/* Local function prototypes */
unsigned char PWMD_DynamicCompress(unsigned char ucValue);

/* Internal logaritmic table */
#ifdef _SOFT_COMPRESSION_
const unsigned char gucPWMDLogVal[32] = 
{
	0,	0,	1,	1,	2,	2,	3,	3,	4,	4,	5,	5,	6,	6,	7,	7,
	8,	8,	9,	10,	11,	11,	11,	12,	12,	13,	14,	14,	14,	15,	15,	15
};
#else
const unsigned char gucPWMDLogVal[32] = 
{
	0,	3,	5,	6,	7,	8,	8,	9,	10,	10,	10,	11,	11,	11,	12,	12,
	12,	13,	13,	13,	13,	13,	14,	14,	14,	14,	14,	14,	15,	15,	15,	15
};
#endif

/* ----------------------------------------------------------- */

/* Some general defines */
#define MAXPWMVAL	0x00
#define MINPWMVAL	0x0F

#define PWMVALMASK	0x0F

/* Power Supply PWM module related defines */
#define PTPERVAL			0x18CD
#define PWMOFF				0x0010
#define PWMON					(PTPERVAL)-1

#define LOWPWMOUT(a)	PDC1 = (a)
#define MIDPWMOUT(a)	PDC2 = (a)
#define HIPWMOUT(a)		PDC3 = (a)

#define PSPWMINIT(regbits,reg) \
	PWMCON##regbits.DTC = 2; 	/* Dead time function disabled */\
	IOCON##regbits.PENL = 1;	/* PWML bit is controlled by PWM subsystem */\
	IOCON##regbits.PMOD = 1;	/* PWM I/O pins are in independent output mode */\
	IOCON##regbits.POLL = 1;	/* PWM_L is low active */\
	FCLCON##regbits.FLTSRC = reg;	/* Source for fault control signal (comparator) */\
	FCLCON##regbits.FLTMOD = 1; /* On Fault condition, set output for the complete cycle */\
	LEBCON##regbits.LEB = 30;	/* Blank fault condition for 240ns */
	
#define CMPINIT(regbits) \
	CMPCON##regbits.INSEL = 0;	/* Select CMPxA input pin */\
	CMPCON##regbits.RANGE = 0;	/* Reference DAC = 1.2V 1% */\
	CMPDAC##regbits.CMREF = 0x12B; /* 350mV for current fault */\
	CMPCON##regbits.CMPON	= 1;	/* Comparator is enabled */

/* ----------------------------------------------------------- */

/* The PWMD subsystem will use TMR2 as main base clock */
/* and the Power Supply PWM module to drive constant current to LEDS */
void PWMD_Init(void)
{
	/* Initialize local variables */
	gtPWMDData.ucLowPwmValue = MINPWMVAL;
	gtPWMDData.ucMidPwmValue = MINPWMVAL;
	gtPWMDData.ucHiPwmValue = MINPWMVAL;
	gtPWMDData.ucBrightTimer = 0;
	gtPWMDData.ucProcess = 0;
	
	/* Initialize Hardware PWM registers */
	PTPER = 0x18CD;			/* Freq = 150kHz -> PTPER = 6,6us/1.05ns */
	PSPWMINIT(1bits,0);
	PSPWMINIT(2bits,1);
	PSPWMINIT(3bits,2);
	
	/* Ensure all channels are turned off */
	LOWPWMOUT(PWMOFF);	
	MIDPWMOUT(PWMOFF);
	HIPWMOUT(PWMOFF);

	/* Initialize the comparators to be used as reference for current limit */
	CMPINIT(1bits);
	CMPINIT(2bits);
	CMPINIT(3bits);
	
	/* Enable Power Supply PWM subsystem */
	PTCONbits.PTEN = 1;
	
	/* Initialize Timer2 registers */
	PR2 = 0x3955;				/* Fref = 125Hz -> Ftic = 125*16 = 2KHz */
	IFS0bits.T2IF = 0;
	IEC0bits.T2IE = 1;
	T2CONbits.TON = 1;
}

/* ----------------------------------------------------------- */

unsigned char PWMD_SetLowValue(unsigned char ucPwmVal)
{
	gtPWMDData.ucLowPwmValue = PWMD_DynamicCompress(ucPwmVal);
	
	return ucPwmVal;
}

/* ----------------------------------------------------------- */

unsigned char PWMD_SetMidValue(unsigned char ucPwmVal)
{
	gtPWMDData.ucMidPwmValue = PWMD_DynamicCompress(ucPwmVal);

	return ucPwmVal;
}

/* ----------------------------------------------------------- */

unsigned char PWMD_SetHiValue(unsigned char ucPwmVal)
{
	gtPWMDData.ucHiPwmValue = PWMD_DynamicCompress(ucPwmVal);

	return ucPwmVal;
}

/* ----------------------------------------------------------- */

/* Dynamically compress the requested PWM in order to adapt
	to this module bit precision. It applies also a logaritmic
	curve to generate a more realistic lights effects */
unsigned char PWMD_DynamicCompress(unsigned char ucValue)
{
	/* Clamp to a 6 bit value */
	if (ucValue >= 32)
		ucValue = 31;

	/* Retrieve the log value */
	ucValue = MINPWMVAL - gucPWMDLogVal[ucValue];

	return ucValue;
}

/* ----------------------------------------------------------- */

/* This is the PWM software generator main loop and it has to 
	 be called periodically by the main program */
void PWMD_MainLoop(void)
{
	/* It's time to process? */
	if (!gtPWMDData.ucProcess)
		return;

	/* Yes... but only one time, please */
	gtPWMDData.ucProcess = 0;

	/* Turn on all channels if required */
	if (gtPWMDData.ucBrightTimer == MINPWMVAL)
	{
		if  (gtPWMDData.ucLowPwmValue != MINPWMVAL)
			LOWPWMOUT(PWMON);
		if (gtPWMDData.ucMidPwmValue != MINPWMVAL)
			MIDPWMOUT(PWMON);
		if (gtPWMDData.ucHiPwmValue != MINPWMVAL)
			HIPWMOUT(PWMON);
	}
	
	/* Test all channels and turn them off if required */
	if (gtPWMDData.ucLowPwmValue > gtPWMDData.ucBrightTimer)
		LOWPWMOUT(PWMOFF);
	if (gtPWMDData.ucMidPwmValue > gtPWMDData.ucBrightTimer)
		MIDPWMOUT(PWMOFF);
	if (gtPWMDData.ucHiPwmValue > gtPWMDData.ucBrightTimer)
		HIPWMOUT(PWMOFF);
}

/* ----------------------------------------------------------- */

/* Timer2 interrupt service routine (ISR) */
void __attribute__((__interrupt__)) _T2Interrupt(void)
{
	/* Clear the pending interrupt */
	IFS0bits.T2IF = 0;
	
	/* Modify local timer and signal the event to the main loop */
	if (!gtPWMDData.ucProcess)
	{
		gtPWMDData.ucBrightTimer--;
		gtPWMDData.ucBrightTimer &= PWMVALMASK;

		gtPWMDData.ucProcess = 1;
	}
}
