/* ----------------------------------------------------------
		
		Psychedelic led lights with dsPic 30F2020 
		Marco Signorini - marco.signorini@libero.it
		
		File: rs232.c
	-----------------------------------------------------------
*/

#include "hardware.h"
#include "rs232.h"

/* Variables definitions ---------------------------------------------------- */

unsigned char gucTxBuffer[TXBUFFLENGTH+1];		/* The Tx Buffer			*/
unsigned char * gpucTxHead;										/* Pointer to the head of Tx 		*/
unsigned char * gpucTxTail;										/* Pointer to the tail of Tx 		*/
																							/* FIFO is empty if Head=Tail 	*/
																							/* FIFO is full  if Head+1=Tail */
																							
/* Internal function prototypes --------------------------------------------- */
unsigned char PutCharInTxBuff(unsigned char cValue);
unsigned char GetCharFromTxBuff(unsigned char * pcValue);

#define UA1_TX_DATA(val)	PutCharInTxBuff((val))

/* ----------------------------------------------------------- */

void RS232_Init(void)
{
	U1MODEbits.ALTIO = 1;		/* Use alternate I/O pins for UART1 */
	U1MODEbits.BRGH = 1;		/* High Speed Mode */
	U1BRG = 0x42;						/* 115200 baud @ 30MHz */
	IFS0bits.U1TXIF = 0;		/* Clear all Tx pending interrupts */
	IEC0bits.U1TXIE = 1;		/* Enable interrupt on Tx */
	U1MODEbits.UARTEN = 1;	/* Enable UART subsystem */

	/* Initialize the pointers */
	gpucTxHead = gucTxBuffer;
	gpucTxTail = gucTxBuffer;

	U1STAbits.UTXEN = 1;
	U1TXREG = 0x00;
}

/* ----------------------------------------------------------- */

unsigned char PutCharInTxBuff(unsigned char cValue)
{
	unsigned char *pucTemp;

	/* Disable Tx interrupt */
	IEC0bits.U1TXIE = 0;
	
	/* Initialize temporary variable */
	pucTemp = gpucTxHead;

	/* Try to put data into the FIFO */
	*(pucTemp++) = cValue;

	/* Need to wrap? */
	if (pucTemp == &(gucTxBuffer[TXBUFFLENGTH]))
		pucTemp = gucTxBuffer;
		
	/* The fifo is full ? */
	if (pucTemp==gpucTxTail)
	{
		/* Enable Tx interrupt before returning */
		IEC0bits.U1TXIE = 1;
		return 0xFF;
	}

	/* OK, update the head pointer */
	gpucTxHead = pucTemp;

	/* Enable Tx interrupt before returning */
	IEC0bits.U1TXIE = 1;

	return 0;
}

/* ----------------------------------------------------------- */

/* Retrieve a char to be transmitted. This has to be called in 
	 the Tx interrupt handler */
unsigned char GetCharFromTxBuff(unsigned char * pcValue)
{
	/* The FIFO is empty */
	if (gpucTxHead == gpucTxTail)
		return 0;

	/* Get the value from the FIFO */
	*pcValue = (char) *(gpucTxTail++);

	/* Need to wrap? */
	if (gpucTxTail == &(gucTxBuffer[TXBUFFLENGTH]))
		gpucTxTail = gucTxBuffer;

	return 1;
}

/* ----------------------------------------------------------- */

/* Dump a char */
void RS232_TxChar(unsigned char ucValue)
{
	UA1_TX_DATA(ucValue);
}

/* ----------------------------------------------------------- */

/* Dump a series of bytes */
void RS232_TxValue(void *pBuffer, unsigned char ucLength)
{
	unsigned char *pData = (unsigned char*) pBuffer;
	
	while (ucLength--)
	{
		UA1_TX_DATA(*pData);
		pData++;
	}
}

/* ----------------------------------------------------------- */

void __attribute__((__interrupt__)) _U1TXInterrupt(void)
{
	unsigned char ucValue;
	
	IFS0bits.U1TXIF = 0;
	if (!GetCharFromTxBuff(&ucValue))
		ucValue = 0x00;

	U1TXREG = ucValue;
}

